/**
 Copyright (c) 2018, Itty Bitty Apps Pty Ltd
 All rights reserved.

 Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 **/

import Foundation
import PlaygroundSupport

/// Helpers for loading and starting Reveal Server framework.
public struct RevealServer {

    public enum Error: Swift.Error {
        /// RevealServer.framework was not found at either of the default locations.
        case frameworkNotFound
        /// Framework binary failed to load.
        case frameworkFailedToLoad
        /// Framework is already loaded.
        case frameworkAlreadyLoaded
        /// Framework has not been loaded yet.
        case frameworkNotLoaded
    }

    // MARK: - Loading

    /// Loads `RevealServer.framework` from one of the default locations: either from playground's `Resources`, or from `Shared Playground Data` directory – optionally starting the server after that.
    /// - parameter autoStart: Whether to start the server automatically upon loading.
    public static func load(autoStart: Bool = true) throws {
        let frameworkURL: URL
        if let url = self.mainBundleFrameworkURL {
            frameworkURL = url
        } else if let url = self.sharedDataFrameworkURL {
            frameworkURL = url
        } else {
            throw Error.frameworkNotFound
        }

        try self.load(from: frameworkURL, autoStart: autoStart)
    }

    /// Loads the Reveal Server framework from the specified URL, optionally starting the server after that.
    /// - parameter frameworkURL: Location of Reveal Server framework to load.
    /// - parameter autoStart: Whether to start the server automatically upon loading.
    public static func load(from frameworkURL: URL, autoStart: Bool = true) throws {
        guard self.loaderClass == nil else {
            throw Error.frameworkAlreadyLoaded
        }

        // Locate the framework's binary
        let revealServerBinaryURL = frameworkURL.appendingPathComponent("RevealServer")
        _ = try revealServerBinaryURL.checkResourceIsReachable()

        // Load the binary
        let revealLoadingResult = dlopen(revealServerBinaryURL.path, RTLD_NOW)
        if revealLoadingResult == nil {
            throw Error.frameworkFailedToLoad
        }

        print("Reveal Server v\(self.version) loaded successfully.")

        if autoStart {
            try self.start()
        }
    }

    // MARK: - Starting/stopping

    /// Starts Reveal Server, enabling inspection of the view hierarchy.
    public static func start() throws {
        guard self.loaderClass != nil else {
            throw Error.frameworkNotLoaded
        }

        NotificationCenter.default.post(name: .revealRequestStart, object: nil)
        print("Reveal Server started.")
    }

    /// Stops Reveal Server.
    public static func stop() {
        NotificationCenter.default.post(name: .revealRequestStop, object: nil)
        print("Reveal Server stopped.")
    }

    // MARK: - Status

    public enum Status {
        /// Reveal Server framework is not loaded yet.
        case notLoaded
        /// Reveal Server is not running.
        case stopped
        /// Reveal Server is running.
        case running
    }

    /// Returns the current status of Reveal Server.
    public static var status: Status {
        guard let serverClass = self.serverClass else {
            return .notLoaded
        }

        let running = serverClass.value(forKeyPath: "sharedServer.httpServer.isRunning") as! NSNumber
        if running.boolValue {
            return .running
        } else {
            return .stopped
        }
    }

    /// Returns the protocol version number of Reveal Server.
    public static var version: Int {
        guard let protocolClass = self.protocolClass else {
            return 0
        }

        return (protocolClass.value(forKey: "protocolVersion") as! NSNumber).intValue
    }

    // MARK: - Private

    private static var mainBundleFrameworkURL: URL? {
        return Bundle.main.url(forResource: "RevealServer", withExtension: "framework")
    }

    private static var sharedDataFrameworkURL: URL? {
        let url = playgroundSharedDataDirectory.appendingPathComponent("RevealServer.framework", isDirectory: true)
        do {
            _ = try url.checkResourceIsReachable()
            return url
        } catch {
            return nil
        }
    }

    private static var loaderClass: AnyClass? {
        return NSClassFromString("IBARevealLoader")
    }

    private static var serverClass: AnyClass? {
        return NSClassFromString("IBARevealServer")
    }

    private static var protocolClass: AnyClass? {
        return NSClassFromString("IBARevealProtocol")
    }

}

private extension Notification.Name {

    static let revealRequestStart = Notification.Name("IBARevealRequestStart")
    static let revealRequestStop = Notification.Name("IBARevealRequestStop")

}
